//============================================================================
//  The contents of this file are covered by the Viskores license. See
//  LICENSE.txt for details.
//
//  By contributing to this file, all contributors agree to the Developer
//  Certificate of Origin Version 1.1 (DCO 1.1) as stated in DCO.txt.
//============================================================================

//============================================================================
//  Copyright (c) Kitware, Inc.
//  All rights reserved.
//  See LICENSE.txt for details.
//
//  This software is distributed WITHOUT ANY WARRANTY; without even
//  the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the above copyright notice for more information.
//============================================================================
#ifndef viskores_cont_serial_DeviceAdapterSerial_h
#define viskores_cont_serial_DeviceAdapterSerial_h

// Keep headers in this order.
// clang-format off
#include <viskores/cont/serial/internal/DeviceAdapterTagSerial.h>
#include <viskores/cont/serial/internal/DeviceAdapterRuntimeDetectorSerial.h>
#include <viskores/cont/serial/internal/RuntimeDeviceConfigurationSerial.h>
#include <viskores/cont/serial/internal/DeviceAdapterMemoryManagerSerial.h>
#include <viskores/cont/serial/internal/DeviceAdapterAlgorithmSerial.h>
// clang-format on

#endif //viskores_cont_serial_DeviceAdapterSerial_h
