//============================================================================
//  Copyright (c) Kitware, Inc.
//  All rights reserved.
//  See LICENSE.txt for details.
//
//  This software is distributed WITHOUT ANY WARRANTY; without even
//  the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the above copyright notice for more information.
//============================================================================
#ifndef viskores_thirdparty_diy_cast_h
#define viskores_thirdparty_diy_cast_h

#include "pre-include.h"
// clang-format off
#include VISKORES_DIY_INCLUDE(mpi/mpi_cast.hpp)
// clang-format on
#include "post-include.h"

#endif // viskores_thirdparty_diy_cast_h
