//============================================================================
//  The contents of this file are covered by the Viskores license. See
//  LICENSE.txt for details.
//
//  By contributing to this file, all contributors agree to the Developer
//  Certificate of Origin Version 1.1 (DCO 1.1) as stated in DCO.txt.
//============================================================================

//============================================================================
//  Copyright (c) Kitware, Inc.
//  All rights reserved.
//  See LICENSE.txt for details.
//
//  This software is distributed WITHOUT ANY WARRANTY; without even
//  the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the above copyright notice for more information.
//============================================================================
// Copyright (c) 2018, The Regents of the University of California, through
// Lawrence Berkeley National Laboratory (subject to receipt of any required approvals
// from the U.S. Dept. of Energy).  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright notice, this
//     list of conditions and the following disclaimer.
//
// (2) Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
// (3) Neither the name of the University of California, Lawrence Berkeley National
//     Laboratory, U.S. Dept. of Energy nor the names of its contributors may be
//     used to endorse or promote products derived from this software without
//     specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
// BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
// OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
// OF THE POSSIBILITY OF SUCH DAMAGE.
//
//=============================================================================
//
//  This code is an extension of the algorithm presented in the paper:
//  Parallel Peak Pruning for Scalable SMP Contour Tree Computation.
//  Hamish Carr, Gunther Weber, Christopher Sewell, and James Ahrens.
//  Proceedings of the IEEE Symposium on Large Data Analysis and Visualization
//  (LDAV), October 2016, Baltimore, Maryland.
//
//  The PPP2 algorithm and software were jointly developed by
//  Hamish Carr (University of Leeds), Gunther H. Weber (LBNL), and
//  Oliver Ruebel (LBNL)
//==============================================================================

#ifndef viskores_filter_scalar_topology_worklet_select_top_volume_branches_binary_search_worklet_h
#define viskores_filter_scalar_topology_worklet_select_top_volume_branches_binary_search_worklet_h

#include <viskores/worklet/WorkletMapField.h>


namespace viskores
{
namespace worklet
{
namespace scalar_topology
{
namespace select_top_volume_branches
{

/// Worklet of a binary search.
/// This is used to check whether the global regular id in inside the block.
/// The global regular Ids in block should be sorted beforehand.
/// This worklet may be reused for many other processes with binary search.
class BinarySearchWorklet : public viskores::worklet::WorkletMapField
{
public:
  using ControlSignature = void(
    FieldIn regularId,         // (input) global regular id
    WholeArrayIn idsArray,     // (array input) all known global regular ids
    FieldOut inBlockIndicator, // (output) 1 if the regularId is inside the idsArray
    FieldOut inBlockIdx        // (output) the index of regularId in idsArray
  );
  using ExecutionSignature = void(_1, _2, _3, _4);
  using InputDomain = _1;

  /// Constructor
  VISKORES_EXEC_CONT
  BinarySearchWorklet() {}

  /// The functor uses binary search to locate regularId in idsArray
  /// if the search fails, return NO_SUCH_ELEMENT; otherwise, return the location
  /// idsArray should be sorted
  template <typename InIdPortalType>
  VISKORES_EXEC void operator()(const viskores::Id& regularId,
                                const InIdPortalType& idsArray,
                                viskores::Id& inBlockIndicator,
                                viskores::Id& inBlockIdx) const
  {
    viskores::Id head = 0;
    viskores::Id tail = idsArray.GetNumberOfValues() - 1;
    inBlockIndicator = 0;
    inBlockIdx = viskores::worklet::contourtree_augmented::NO_SUCH_ELEMENT;

    while (head <= tail)
    {
      viskores::Id mid = (head + tail) >> 1;
      viskores::Id midValue = idsArray.Get(mid);
      if (regularId == midValue)
      {
        inBlockIdx = mid;
        inBlockIndicator = 1;
        return;
      }
      else if (regularId > midValue)
        head = mid + 1;
      else
        tail = mid - 1;
    }
  }
}; // BinarySearchWorklet

} // namespace select_top_volume_branches
} // namespace scalar_topology
} // namespace worklet
} // namespace viskores

#endif
